/*
 * Copyright 2012 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License version 3, as published 
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranties of 
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along 
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *      David King <david.king@canonical.com>
 */

/**
 * Web credentials preferences widget. This can be packed into the GNOME
 * Control Center or used in a standalone application.
 */
public class Cc.Credentials.Preferences : Gtk.Notebook
{
    private AuthorizationPage authorization_page;
    private Ag.Manager accounts_manager;

    /* This must be a construct property so that is is called before the
     * construct block.
     */
    public uint account_details_id { get; construct; }
    public string application_id { get; construct; }

    /**
     * Select the desired page of @link Preferences to query the current
     * accounts and providers, or to authorize an account.
     *
     * @param ACCOUNTS select an account or provider in a two-pane view
     * @param AUTHORIZATION authorize an account in a full-size widget, such as
     * a WebKit view.
     */
    private enum PreferencesPage
    {
        ACCOUNTS = 0,
        AUTHORIZATION = 1,
    }

    public Preferences ()
    {
        Object ();
    }

    public Preferences.with_account_details (Ag.AccountId account_id)
    {
        Object (account_details_id: account_id);
    }

    public Preferences.with_application (string application)
    {
        Object (application_id: application);
    }

    construct
    {
        show_tabs = false;
        show_border = false;
        expand = true;
        border_width = 12;

        accounts_manager = new Ag.Manager ();

        AccountsPage accounts_page;

        if (account_details_id != 0)
        {
            accounts_page = new AccountsPage.with_account_details (account_details_id);
        }
        else if (application_id != null)
        {
            accounts_page = new AccountsPage.with_application (application_id);
        }
        else
        {
            accounts_page = new AccountsPage ();
        }

        accounts_page.new_account_request.connect (on_accounts_page_new_account_request);
        accounts_page.reauthenticate_account_request.connect (on_accounts_page_reauthenticate_account_request);

        this.append_page (accounts_page);

        authorization_page = new AuthorizationPage ();
        authorization_page.cancelled.connect (on_authorization_page_cancelled);

        this.append_page (authorization_page);

        set_current_page (PreferencesPage.ACCOUNTS);

        set_size_request (-1, 400);

        show ();
    }

    /**
     * Handle the new-account-request signal from AccountsPage (and in turn
     * ProvidersPage), switching notebook page to the new account view.
     *
     * @param provider_name the name of the provider for which to add an
     * account
     */
    private void on_accounts_page_new_account_request (string provider_name)
    {
        var account = accounts_manager.create_account (provider_name);
        authorization_page.account = account;
        set_current_page (PreferencesPage.AUTHORIZATION);
    }

    /**
     * Handle the reauthenticate-account-request signal from AccountsPage (and
     * in turn AccountDetailsPage), switching notebook page to the account
     * authentication view.
     *
     * @param account the account to authenticate
     */
    private void on_accounts_page_reauthenticate_account_request (Ag.Account account)
    {
        authorization_page.reauthenticate_account (account);
        set_current_page (PreferencesPage.AUTHORIZATION);
    }

    /**
     * Handle the authorization process for a new account being cancelled,
     * switching the current notebook page to the provider selection view.
     */
    private void on_authorization_page_cancelled ()
    {
        set_current_page (PreferencesPage.ACCOUNTS);
    }
}
