/*
 * Copyright 2012 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License version 3, as published 
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranties of 
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
 * PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along 
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *      David King <david.king@canonical.com>
 */

/**
 * Web credentials providers page. This is for selecting an available provider
 * to add a new account for.
 */
public class Cc.Credentials.ProvidersPage : Gtk.Grid
{
    /**
     * Emitted when a new account should be added.
     *
     * @param provider_name the name of the provider for which an account
     * should be added
     */
    public signal void new_account_request (string provider_name);

    private Ag.Manager manager;
    private Gtk.ComboBox applications_combo;
    private Gtk.TreeModelFilter filter_model;
    private string current_application;

    public string application_id { get; construct; }

    public ProvidersPage ()
    {
        Object ();
    }

    public ProvidersPage.with_application (string application)
    {
        Object (application_id: application);
    }

    construct
    {
        orientation = Gtk.Orientation.VERTICAL;
        expand = true;

        // TODO: Use common Ag.Manager.
        manager = new Ag.Manager ();

        this.add (create_providers_infobar ());
        this.add (create_providers_tree ());

        /* This defaults to -1, no selection, so force it to have the first
         * item selected.
         */
        applications_combo.active = 0;

        // Trigger a refilter if an application was passed in.
        if (application_id != null)
        {
            current_application = application_id;

            Gtk.TreeIter iter;
            var applications_model = applications_combo.model as ApplicationsModel;
            if (applications_model.find_iter_for_application (application_id,
                                                              out iter))
            {
                applications_combo.set_active_iter (iter);
            }
            else
            {
                message ("Passed-in application '%s' was not found",
                         application_id);
            }
        }

        set_size_request (-1, 400);

        show ();
    }

    /**
     * Create the infobar which describes the treeview of account providers.
     *
     * @return a Gtk.InfoBar describing the account providers
     */
    private Gtk.Widget create_providers_infobar ()
    {
        var infobar = new Gtk.InfoBar ();
        infobar.message_type = Gtk.MessageType.QUESTION;
        var label = new Gtk.Label (_("Show accounts that integrate with:"));
        var applications_model = new ApplicationsModel ();
        applications_combo = new Gtk.ComboBox.with_model (applications_model);
        applications_combo.hexpand = true;
        var text_renderer = new Gtk.CellRendererText ();
        applications_combo.pack_start (text_renderer, false);
        applications_combo.set_attributes (text_renderer, "markup",
                                           ApplicationsModel.ModelColumns.APPLICATION_DESCRIPTION);
        var infobar_grid = new Gtk.Grid ();
        infobar_grid.column_spacing = 6;
        infobar_grid.add (label);
        infobar_grid.add (applications_combo);
        var content_area = infobar.get_content_area () as Gtk.Container;
        content_area.add (infobar_grid);

        applications_combo.changed.connect (on_applications_combo_changed);

        infobar.show_all ();

        return infobar;
    }

    /**
     * Create the treeview of account providers, for setting up a new account.
     *
     * @return a Gtk.ScrolledWindow containing the account provider treeview
     */
    private Gtk.Widget create_providers_tree ()
    {
        var providers_model = new ProvidersModel ();
        filter_model = new Gtk.TreeModelFilter (providers_model, null);
        filter_model.set_visible_func (filter_model_visible);
        var providers_tree = new Gtk.TreeView.with_model (filter_model);
        providers_tree.headers_visible = false;
        var pixbuf_renderer = new Gtk.CellRendererPixbuf ();
        pixbuf_renderer.stock_size = Gtk.IconSize.DIALOG;
        pixbuf_renderer.set_padding (6, 6);
        var text_renderer = new Gtk.CellRendererText ();
        var providers_tree_column = new Gtk.TreeViewColumn ();
        providers_tree_column.pack_start (pixbuf_renderer, false);
        providers_tree_column.add_attribute (pixbuf_renderer, "gicon",
                                             ProvidersModel.ModelColumns.PROVIDER_ICON);
        providers_tree_column.pack_start (text_renderer, true);
        providers_tree_column.add_attribute (text_renderer, "markup",
                                             ProvidersModel.ModelColumns.PROVIDER_DESCRIPTION);
        providers_tree.append_column (providers_tree_column);

        var scrolled_window = new Gtk.ScrolledWindow (null, null);
        scrolled_window.min_content_height = 250;
        scrolled_window.min_content_width = 200;
        scrolled_window.expand = true;
        scrolled_window.shadow_type = Gtk.ShadowType.ETCHED_IN;
        scrolled_window.add (providers_tree);

        var providers_selection = providers_tree.get_selection ();
        providers_selection.changed.connect (on_providers_selection_changed);

        scrolled_window.show_all ();

        return scrolled_window;
    }

    /**
     * Visible filter function for the filter tree model of applications.
     *
     * @param model the child tree model to filter
     * @param iter the iter to query data from
     */
    private bool filter_model_visible (Gtk.TreeModel model, Gtk.TreeIter iter)
    {
        string application_name;
        model.get (iter,
                   ApplicationsModel.ModelColumns.APPLICATION_NAME, out application_name,
                   -1);

        return application_name == current_application;
    }

    /**
     * Handle the application combo box selection being changed, by updating
     * the current application being filtered by the filter model.
     *
     * @param applications_combo the combo box for selecting from a list of
     * applications
     */
    private void on_applications_combo_changed (Gtk.ComboBox applications_combo)
    {
        Gtk.TreeIter iter;
        if (applications_combo.get_active_iter (out iter))
        {
            string application_name;
            var model = applications_combo.model;
            model.get (iter,
                       ApplicationsModel.ModelColumns.APPLICATION_NAME, out application_name,
                       -1);

            current_application = application_name;
            filter_model.refilter ();
        }
    }

    /**
     * Tell the main panel to show a new account page for the selected
     * provider.
     *
     * @param selection the selection of the providers treeview
     */
    private void on_providers_selection_changed (Gtk.TreeSelection selection)
    {
        Gtk.TreeModel model;
        Gtk.TreeIter iter;

        if (selection.get_selected (out model, out iter))
        {
            string provider_name;
            model.get (iter,
                       ProvidersModel.ModelColumns.PROVIDER_NAME, out provider_name,
                       -1);

            /* Unselect the row, as otherwise it is impossible to add another
             * account from the same provider immediately afterwards.
             */
            selection.unselect_all ();

            // Emit signal for the main panel to switch notebook page.
            new_account_request (provider_name);
        }
    }
}
